#
# Virtual Traces Experiment Software Distribution.
# This code is licensed under GPL v2, and is copyright (C) Jack Whitham 2008.
# See PATCH_NOTE.
#

# build/ALPHA_SE/m5.debug --trace-flags=Fetch,Exec configs/example/se.py -c ./compress.elf >tlog


import pickle, sys

MAX_META_INST_LENGTH = 20
COND_BRANCH_SET = set([])
JUMP_SET = set([])
BOUNDARY_TYPES = [ NOT_TAKEN, TAKEN, JUMP, FALLTHROUGH ] = [ 
            "Not_taken", "Taken", "Jump", "Fallthrough" ]
BRANCH_NOTIFY = "Branch=["
COST_TABLE = {
    "IntAlu" : 1,
    "IntMult" : 3,
    "IntDiv" : 20,
    "MemRead" : 1,
    "MemWrite" : 1,
    "No_OpClass" : 1,
    "FloatAdd" : 2,
    }
MIN_COST = 1

def Get_Op_Cost(fu_type):
    return COST_TABLE[ fu_type ]


class Branch_Record:
    def __init__(self, src_pc, target_pc):
        self.src_pc = src_pc
        self.target_pc = target_pc
        self.entry = False
        self.conditional = False
        self.taken = JUMP
        self.info = 'Unknown'

    def Type_Is_Conditional(self, taken):
        assert type(taken) == bool
        self.conditional = True
        self.taken = BOUNDARY_TYPES[ int(taken) ]
        self.info = 'Conditional %s' % self.taken

    def Type_Is_Call(self):
        assert not self.conditional
        self.taken = JUMP
        self.info = 'Call'

    def Type_Is_Return(self):
        assert not self.conditional
        self.taken = JUMP
        self.info = 'Return'

    def Type_Is_Entry(self):
        assert not self.conditional
        self.entry = True
        self.info = 'Entry'

    def Add_Hint(self, hint):
        self.info = '%s: %s' % (hint, self.info)

def Capture_Branch_Data(exec_trace_file):
    live_code = dict()
    branch_data = []
    branch_is_exit = False
    branch_detect = False
    branch_ignore = False
    serialising_instructions = set()
    address_hint = None
    opcode = None
    prev_line = None
    pc = npc = None
    for line in exec_trace_file:
        fields = line.split(':')
        if ( len(fields) < 6 ):
            continue
        if ( not ( fields[ 1 ].startswith(' system.cpu') )):
            continue

        if ( fields[ 2 ].endswith('Fetch') 
        and fields[ 3 ].endswith('PC') ):
            # fetch
            last_pc = pc
            last_npc = npc

            pc = int(fields[ 4 ].split()[ 0 ], 16)
            npc = int(fields[ 5 ].split()[ 0 ], 16)
            if (( pc != last_pc ) or ( npc != last_npc )):
                branch_taken = (( pc != last_npc ) or branch_ignore )
                br = None


                if ( branch_is_exit ):
                    serialising_instructions.add(last_pc)
                    br = Branch_Record(last_pc, pc)
                    br.Type_Is_Conditional(branch_taken)
                elif ( branch_detect ):
                    # Conditional branch
                    br = Branch_Record(last_pc, pc)
                    br.Type_Is_Conditional(branch_taken)
                elif ( branch_taken ):
                    br = Branch_Record(last_pc, pc)
                    if ( not branch_ignore 
                    and ( prev_line != None )):
                        print 'Unexpectedly caused a jump:'
                        print prev_line
                        assert False
                        return None
                    elif ( last_pc != None ):
                        # Unconditional branch, call, or return
                        br = Branch_Record(last_pc, pc)
                        # What type is it?
                        if ( opcode.startswith("bl") ):
                            br.Type_Is_Call()
                        elif ( opcode.startswith('bx') ):
                            br.Type_Is_Return()
                    else:
                        # Program entry
                        br = Branch_Record(pc, npc)
                        br.Type_Is_Entry()

                if ( br != None ):
                    if ( address_hint != None ):
                        br.Add_Hint(address_hint)
                        address_hint = None

                    branch_data.append(br)

                branch_detect = False
                branch_ignore = False
                branch_is_exit = False

        elif ( fields[ 1 ].endswith('T0 ') ):
            # execute
            opcode_fields = fields[ 3 ].split()
            opcode = opcode_fields[ 0 ]
            bn = line.find(BRANCH_NOTIFY)
            if ( bn > 0 ):
                bn += len(BRANCH_NOTIFY)
                branch_flags = line[ bn : -1 ]
                if ( "S" in branch_flags ):
                    branch_is_exit = True
                elif ( "C" in branch_flags ):
                    branch_detect = True
                else:
                    branch_ignore = True

                if ( address_hint != None ): # ever used?
                    address_hint = fields[ 2 ]

            fu_type = fields[ 4 ].strip()
            live_code[ pc ] = Get_Op_Cost(fu_type)
            
            prev_line = line

    serialising_instructions = list(serialising_instructions)
    serialising_instructions.sort()

    assert not ( live_code.has_key(None) )
    return (live_code, branch_data, serialising_instructions)

def Print_Branch_Data(branch_data):
    for br in branch_data:
        print '  %010x -> %010x %s' % (br.src_pc, br.target_pc, br.taken)

class Metainstruction:
    def __init__(self, start_pc):
        self.start_pc = start_pc
        self.predictions = []
        self.cost = 0
        self.pcost = 0

    def Add_Prediction(self, taken, target_pc):
        assert taken in BOUNDARY_TYPES
        if ( taken != FALLTHROUGH ):
            print 'prediction taken %s target %x cost %u' % (
                        taken[ 0 ], target_pc, self.cost)
            self.pcost = self.cost
            self.predictions.append((taken, target_pc, self.cost))

    def Add_Cost(self, bb):
        self.cost += bb.Get_Cost()

    def Get_Length(self):
        return len(self.predictions)

    def Get_Pred(self, index):
        try:
            return self.predictions[ index ]
        except IndexError:
            raise Exception("pc %x has %u predictions - asked for %u" % (
                self.start_pc, len(self.predictions), index))

    def Get_Target_PC(self, index):
        (taken, target_pc, cost) = self.Get_Pred(index)
        return target_pc

    def Get_Taken(self, index):
        (taken, target_pc, cost) = self.Get_Pred(index)
        return taken

    def Get_Cost(self, index):
        if ( self.Get_Length() == 0 ):
            cost = self.cost
        elif ( index >= self.Get_Length() ):
            cost = self.pcost
        else:
            (taken, target_pc, cost) = self.Get_Pred(index)

        return cost

    def Get_Path_Item(self):
        return 'VTS %x\n' % (self.start_pc)

    def Output(self, fd):
        fd.write('STA %x\nLEN %x\n' % (self.start_pc, 
                            len(self.predictions)))
        for (taken, pc, cost) in self.predictions:
            assert taken in BOUNDARY_TYPES
            assert taken != FALLTHROUGH

            if ( taken == JUMP ):
                fd.write('JMP %x\n' % pc)
            elif ( taken == TAKEN ):
                fd.write('BTK %x\n' % pc)
            else:
                fd.write('BNT %x\n' % pc)

class BB_Record:
    def __init__(self, start_pc):
        self.start_pc = start_pc
        self.exit_type = FALLTHROUGH
        self.jump_targets = []
        self.taken_target = None
        self.not_taken_target = None
        self.fallthrough_target = None
        self.jump_counts = []
        self.taken_count = 0
        self.not_taken_count = 0
        self.meta_inst = []
        self.cost = 0

    def Branch(self, target_bb, taken):
        assert isinstance(target_bb, BB_Record)
        assert taken in BOUNDARY_TYPES
        inc_count_by = 1

        if ( taken == JUMP ):
            # Unconditional: jsr, bsr, ret, etc.
            assert self.exit_type in (FALLTHROUGH, JUMP)
            self.exit_type = JUMP

            # Already seen this jump target?
            for i in xrange(len(self.jump_targets)):
                if ( self.jump_targets[ i ] == target_bb ):
                    # Yes
                    self.jump_counts[ i ] += inc_count_by
                    return

            # No
            self.jump_targets.append(target_bb)
            self.jump_counts.append(inc_count_by)

        elif ( taken == NOT_TAKEN ):
            assert self.not_taken_target in (None, target_bb)
            assert self.exit_type != JUMP

            self.not_taken_count += 1
            self.not_taken_target = target_bb
            if ( self.not_taken_count > self.taken_count ):
                self.exit_type = NOT_TAKEN

        elif ( taken == TAKEN ):
            assert self.taken_target in (None, target_bb)
            assert self.exit_type != JUMP

            self.taken_count += 1
            self.taken_target = target_bb
            if ( self.taken_count > self.not_taken_count ):
                self.exit_type = TAKEN

        else:
            assert False, ( "Fallthrough to bb %x?" % target_bb.start_pc )

    def Get_Exit_Type(self):
        return self.exit_type

    def Get_Meta_Inst(self):
        return self.meta_inst

    def Get_Next(self):
        if ( self.exit_type == FALLTHROUGH ):
            #print 'FT'
            return self.fallthrough_target
        elif ( self.exit_type == NOT_TAKEN ):
            #print 'NT'
            return self.not_taken_target
        elif ( self.exit_type == TAKEN ):
            #print 'T'
            return self.taken_target
        elif ( self.exit_type == JUMP ):
            #print 'J'
            count = -1
            target = None
            for i in xrange(len(self.jump_targets)):
                if ( self.jump_counts[ i ] > count ):
                    count = self.jump_counts[ i ]
                    target = self.jump_targets[ i ]
            assert target != None
            return target
        else:
            assert False

    def Get_Branch(self, direction):
        assert self.exit_type in (NOT_TAKEN, TAKEN)
        if ( direction == NOT_TAKEN ):
            return self.not_taken_target
        elif ( direction == TAKEN ):
            return self.taken_target
        else:
            assert False

    def Get_Cost(self):
        return self.cost

    def Set_Cost(self, cost):
        self.cost = cost

    def Fall_Through(self, target_bb):
        self.fallthrough_target = target_bb

    def Follow(self):
        return (self.Get_Next(), self.exit_type)

    def Make_Metainstruction(self):
        #print 'This is bb %x' % self.start_pc
        mi = self.meta_inst = Metainstruction(self.start_pc)
        bb = self
        #print 'bb id', id(bb)
        mi.Add_Cost(bb)
        (bb, taken) = bb.Follow()
        while (( mi.Get_Length() < MAX_META_INST_LENGTH )
        and ( bb != self )
        and ( bb != None )):
            #print '%u: _ %x %s' % (mi.Get_Length(), bb.start_pc, taken)
            mi.Add_Prediction(taken, bb.start_pc)
            mi.Add_Cost(bb)
            (bb, taken) = bb.Follow()
            #print 'bb id', id(bb), ( bb == self ), bb
        
        loop_size = mi.Get_Length()

        if (( bb == self )
        and ( mi.Get_Length() < ( MAX_META_INST_LENGTH - 1 - loop_size ))):
            # Loop detected. Only finish at the entrance
            #print '%u: %x %s' % (mi.Get_Length(), bb.start_pc, taken)
            while ( mi.Get_Length() < ( MAX_META_INST_LENGTH - 1 - loop_size )):

                self.meta_inst.Add_Prediction(taken, bb.start_pc)
                mi.Add_Cost(bb)
                (bb, taken) = bb.Follow()

                while ( bb != self ):
                    #print '%u: %x %s' % (mi.Get_Length(), bb.start_pc, taken)
                    self.meta_inst.Add_Prediction(taken, bb.start_pc)
                    mi.Add_Cost(bb)
                    (bb, taken) = bb.Follow()

                #print '%u: %x %s' % (mi.Get_Length(), bb.start_pc, taken)

            assert bb == self
            assert mi.Get_Length() < MAX_META_INST_LENGTH
        
        if ( mi.Get_Length() == 0 ):
            self.meta_inst.Add_Prediction(taken, bb.start_pc)
        
    def Get_Path_Item(self):
        return 'BBS %x\n' % (self.start_pc)


def Analyse_Branch_Data(live_code, branch_data, serialising_instructions):

    # Identify basic blocks
    lowest_pc = 1 << 64
    basic_blocks_d = dict()
    basic_blocks_l = []

    def Make_BB(pc):
        if ( live_code.has_key(pc) ):
            if ( basic_blocks_d.get(pc, None) == None ):
                basic_blocks_d[ pc ] = bb = BB_Record(pc)
                basic_blocks_l.append(bb)

    for br in branch_data:
        Make_BB(br.src_pc + 4)
        Make_BB(br.target_pc)
        
    print '%u bbs identified' % len(basic_blocks_d)
    assert len(basic_blocks_d) == len(basic_blocks_l)

    # Special handling for initial BB
    br = branch_data[ 0 ]
    assert br.entry
    Make_BB(br.src_pc)
    start_bb = bb = basic_blocks_d[ br.src_pc ]

    # Initial BB is removed: src_pc and target_pc are unusual
    branch_data_minus_start = branch_data[ 1: ]

    # Sort into address order
    basic_blocks_l.sort(key=lambda bb: bb.start_pc)
    lowest_pc = basic_blocks_l[ 0 ].start_pc

    # Useful function
    def Search_BB(pc):
        while ( basic_blocks_d.get(pc, None) == None ):
            pc -= 4
            assert pc >= lowest_pc, ( "pc %x, lowest pc %x" % (pc, lowest_pc))
        return basic_blocks_d[ pc ]

    # Link branches to target basic blocks. Mark targets.
    for br in branch_data_minus_start:
        src_bb = Search_BB(br.src_pc)
        target_bb = basic_blocks_d[ br.target_pc ]
        src_bb.Branch(target_bb, br.taken)

    # Add fallthroughs
    # Add execution costs to BBs
    bb0 = basic_blocks_l[ 0 ]
    for bb1 in basic_blocks_l[ 1: ]:
        #print 'bb'
        bb0.Fall_Through(bb1)
        cost = 0
        pc = bb0.start_pc
        while ( pc < bb1.start_pc ):
            cost += live_code.get(pc, MIN_COST)
            #print 'pc %x cost %u' % (pc, cost)
            pc += 4

        bb0.Set_Cost(cost)
        bb0 = bb1
    
    # Every bb needs a metainstruction.
    for bb in basic_blocks_l:
        bb.Make_Metainstruction()

    # Make a list of metainstructions, and test them
    meta_insts = []
    end_reached = False
    for bb in basic_blocks_l:
        meta_insts.append(bb.Get_Meta_Inst())

#        # Test - does the metainstruction track
#        # the execution path correctly?
#        # Test - coverage
#        bb_track = bb
#        for taken in bb.meta_inst:
#
#            assert type(taken) == bool
#            assert bb_track != None
#            while ( bb_track.unconditional ):
#                bb_track.is_covered = True
#                bb_track = bb_track.Get_Next()
#                assert bb_track != None
#
#            assert ( bb_track.tcount[ int(not taken) ] 
#                        <= bb_track.tcount[ int(taken) ] )
#            bb_track.is_covered = True
#            bb_track = bb_track.targets[ taken ]
#
#
#        # Test - end of path has another metainstruction,
#        # or end of program.
#        if ( bb_track == None ):
#            assert not end_reached
#            end_reached = True
#        #assert len(bb_track.meta_inst) != 0, hex(bb_track.start_pc)

            
    # (1) Test the effect of the metainstructions
    # on the program trace
    # (2) Record the path through the meta instructions
    hit = miss = 0
    bb_track = start_bb
    meta_inst = start_bb.Get_Meta_Inst()
    meta_inst_ptr = 0
    meta_inst_path = [ meta_inst.Get_Path_Item() ]

    for br in branch_data:

        et = bb_track.Get_Exit_Type() 
        #print '%x->%x mi %x:%u  (%s)' % (br.src_pc, br.target_pc,
        #        abs(id(meta_inst)), meta_inst_ptr, et)

        target_pc = meta_inst.Get_Target_PC(meta_inst_ptr)
        prediction = meta_inst.Get_Taken(meta_inst_ptr) 
        change = False

        if ( et in (FALLTHROUGH, JUMP) ):

            bb_track = bb_track.Get_Next()

            if ( et == JUMP ):
                assert prediction == JUMP
                assert bb_track.start_pc == target_pc

            if ( bb_track.start_pc != br.target_pc ):
                miss += 1
                #print 'Measure: mispredict at %u.' % meta_inst_ptr
                #print 'JSR/Ret: %x predicted, actually %x. Miss %u!' % (
                #            bb_track.start_pc, br.target_pc, miss)
                change = True
            else:
                hit += 1
                assert bb_track != None 

        elif ( et in (NOT_TAKEN, TAKEN) ):
            assert prediction in (NOT_TAKEN, TAKEN)

            bb_track = bb_track.Get_Branch(prediction)

            assert bb_track.start_pc == target_pc
            assert br.taken != None

            if ( br.taken != prediction ):
                miss += 1
                print 'Measure: mispredict at %u.' % meta_inst_ptr
                print 'Branch: %x(%s) predicted, actually %x(%s). Miss %u!' % (
                            bb_track.start_pc, 
                            prediction,
                            br.target_pc, 
                            br.taken,
                            miss)
                change = True
            else:
                hit += 1
                assert bb_track.start_pc == br.target_pc
                assert bb_track != None

        if ( et != FALLTHROUGH ):
            meta_inst_path.append(bb_track.Get_Path_Item())
            meta_inst_ptr += 1
            if ( meta_inst_ptr == meta_inst.Get_Length() ):
                #print 'Measure: no mispredict.'
                change = True

        if ( change ):
            bb_track = basic_blocks_d[ br.target_pc ]
            meta_inst_ptr = 0
            meta_inst = bb_track.Get_Meta_Inst()
            meta_inst_path.append(meta_inst.Get_Path_Item())
            #print 'Issuing MI. pc=%x' % br.target_pc

        assert bb_track.start_pc == br.target_pc
    
    print 'total miss', miss
    print 'total hit', hit
    print 'meta instructions', len(meta_insts)

    return (meta_insts, meta_inst_path)

def Get_Meta_Instructions(tlog_input):
    (live_code, branch_data, 
        serialising_instructions) = Capture_Branch_Data(file(tlog_input))
    return Analyse_Branch_Data(live_code, branch_data, serialising_instructions)

def Process(tlog_input='tlog', output='meta.txt', path_output='path.txt'):
    (meta_insts, meta_inst_path) = Get_Meta_Instructions(tlog_input)

    f = file(output, 'wt')
    for meta_inst in meta_insts:
        meta_inst.Output(f)
    f.close()
    f = file(path_output, 'wt')
    for item in meta_inst_path:
        f.write(item)
    f.close()

def Compare_IIO_O3VTC(tlog_input, meta_input, compare_output):
    (meta_insts, meta_inst_path) = Get_Meta_Instructions(tlog_input)

    meta_inst_exits = dict()
    meta_inst_index = dict()
    for mi in meta_insts:
        meta_inst_index[ mi.start_pc ] = mi

    events = set([ 'JMP', 'BTK', 'BNT', 'END' ])
    mi = None
    mi_exits = None
    index = 0
    out = file(compare_output, 'wt')
    for line in file(meta_input):
        line = line.rstrip()
        fields = line.split()
        cmd = fields[ 0 ]
        param = int(fields[ -1 ], 16)

        if ( cmd == 'STA' ):
            mi = meta_inst_index[ param ]
            out.write('%s\n' % (line))
            mi_exits = meta_inst_exits[ param ] = []
            index = 0
        elif ( cmd in events ):
            iio_cost = mi.Get_Cost(len(mi_exits))
            o3vtc_cost = param
            mi_exits.append((o3vtc_cost, iio_cost))
            out.write('%s %x\n' % (line, iio_cost))
        else:
            out.write('%s\n' % (line))

    return meta_inst_exits

    
